# -------------------------------------------------------------------------------------------------------------------
# D. Cordier - 27 janvier 2023 - Méthode de Lorenz (1993) permettant de calculer la vitesse limite de chute de 
#   gouttes de pluie.
#   Cette méthode a d'abord été implémentée/mise au point dans le notebook 'Dmax_D0.ipynb' avant d'être "versée" 
#   dans ce fichier.
# http://orcid.org/0000-0003-4515-6271
# daniel.cordier@cnrs.fr
#
#  [10.5281/zenodo.10391618](https://zenodo.org/uploads/10391618)
# Provided under GPL licence: https://www.gnu.org/licenses/licenses.fr.html#GPL
#
import numpy as np

# -------------------------------------------------------------------------------------------------------------------
# Viscosité dynamique pour un gaz parfait :
def visco_gp(Mmol,T,d0):
    """
    Calcul la viscosité dynamique d'un gaz parfait.
    Inputs:
      - Mmol ----: masse molaire de l'espèce considée (kg mol^-1).
      - T -------: température absolue (K).
      - d0 ------: ''taille'' de la molécule en question (m).
    Outputs:
      - eta_gp --: viscosité dynamique du gaz considéré (Pa s).
    """
    kB   = 1.38e-23 # J.K^-1
    Navo = 6.02e23  # mol^-1
    m    = Mmol/Navo
    eta_gp= np.sqrt(4*kB*m*T/9/np.pi**3/d0**4)
    
    return eta_gp

# -------------------------------------------------------------------------------------------------------------------
# Viscosité dynamique de N2, repris sur Lorenz (1993) :
def visco_N2_lorenz(T):
    """
    Repris sur Lorenz (1993) Eq. A1 p. 655.
    """
    eta_N2 = 1.718e-5 + 5.1e-8 * (T-273)
    
    return eta_N2

# -------------------------------------------------------------------------------------------------------------------
# Calcul du nombre de Reynolds dans l'air :
def Re_air(rho_air,d,Uair,eta_air):
    """
    Inputs:
      - rho_air: air density (kg m^-3)
      - d: diameter of the droplet (m)
      - Uair: velocity in the air (m s^-1)
      - eta_air: dynamic viscosity of the air (Pa s)
    Outputs:
      - Re_liq: Reynolds number in the droplet.
    """
    return rho_air*d*Uair/eta_air

# -------------------------------------------------------------------------------------------------------------------
# Coefficients de traînée utilisés par Lorenz (1993) :
def Cd_lorenz(Re):
    """
    Voir Eq. 4 de Lorenz (1993)
    Inputs:
      - Re ---: nombre de Reynolds de l'écoulement d'air autour de la goutte.
    Outputs:
      - Cd ---: coefficient de traînée donné par l'Eq. 4 de Lorenz (1993)
    """
    # Voir aussi Eq. 3 p. 111 de Clift et al. (le livre sur les gouttes et bulles) 
    Cd = 24/Re * (1+0.197*Re**0.63+2.6e-4*Re**1.38)
    
    return Cd

def Cd_Clift_Gauvin(Re):
    """
    Inputs:
      - Re ---: nombre de Reynolds de l'écoulement d'air autour de la goutte.
    Outputs:
      - Cd ---: coefficient de traînée donné par l'Eq. 10 Tab. 5.1 p. 111 Clift.
    """
    Cd = 24/Re * (1+0.15*Re**0.687) + 0.42 /(1+4.25e4*Re**(-1.16))
    
    return Cd

# -------------------------------------------------------------------------------------------------------------------
# Nombre de Bond utilisé par Lorenz (1993) :
def Bond_Lorenz(rho_liq,d0,g,sigma):
    """
    Voir équation dans le texte p. 649 de Lorenz (1993).
    Remarque : Lorenz n'utilise pas le diamètre d0 de la goutte mais le rayon a0.
    """
    a0= d0/2
    Bond = rho_liq*a0**2*g / sigma
    return Bond

# -------------------------------------------------------------------------------------------------------------------
# cf. Lorenz (1993) :
def k_Green75(Bond):
    """
    Voir Eq. (13) de Lorenz (1993).
    """
    k = (4/17 * (17*Bond/4 + 1)**(1/2) + 13/17)**(-3/2)
    
    return k

# -------------------------------------------------------------------------------------------------------------------
def F_lorenz_tobe_solved(rho_liq,sigma_liq,rho_air,visco,Mair,T,dim_molecule,d0,g,V,rigid,CdEq):
    """
    Equation à résoudre numériquement pour obtenir la vitesse limite de chute d'une goutte de liquide
    déformée par les forces aérodynamiques (cf. Lorenz, 1993).
    Inputs:
      - rho_liq ------: masse volumique du liquide considéré (kg m^-3).
      - sigma_liq ----: tension superficielle du liquide considéré (N m^-1)
      - rho_air ------: masse volumique de l'air (kg m^-3).
      - visco --------: keyword
                        'GP' ---: the viscosity is given by the law relevant for ideal gases.
                        'LOR' --: the visocity is given by the law use by Lorenz (1993) relevant for N2.
      - Mair ---------: masse molaire des molécules de l'air (kg mol^-1).
      - T ------------: température absolue (K).
      - dim_molecule -: taille des molécules d'air (m).
      - d0 -----------: diamètre de la goutte de liquide, non déformée (m).
      - g ------------: accélération de la pesanteur (m s^-2).
      - V ------------: la vitesse de chute (m s^-1).
      - rigid --------:  =0 : gouttes déformées par rapport à la sphère.
                         =1 : gouttes NON-déformées par rapport à la sphère.
      - CdEq ---------: type of Cd equation used (cf. Clift p. 111)
    Outputs:
      - Fv -----------: l'équation à résoudre F(V)= V^2 - 2Mg/rho_air/S/Cd
    """
    a0= d0/2 # Conversion diamètre -> rayon.
    S0= np.pi * a0**2
    
    # -----------------------------------------------------------------
    # Air viscosity model: --------------------------------------------
    if visco == 'GP':
        eta_air = visco_gp(Mmol=Mair,T=T,d0=dim_molecule)
    else:
        if visco == 'LOR':
            eta_air = visco_N2_lorenz(T)
        else:
            print (" >>>> WARNING: in function 'F_lorenz_tobe_solved' not relevant viscosity option!!!")
    # -----------------------------------------------------------------
    
    Reair   = Re_air(rho_air,d0,V,eta_air)

    if CdEq == 3:
        Cd0 = Cd_lorenz(Reair)
    if CdEq == 10:
        Cd0 = Cd_Clift_Gauvin(Reair) # Cf. Eq 10 Tab. 5.1 p. 111 Clift
        
    if rigid:
        S = S0  
        Cd= Cd0 
    else:
        #We_air  = a0*rho_air*V**2/sigma_liq
        #k= k_ratio_lorenz(We_air)
        
        Bond = Bond_Lorenz(rho_liq,d0,g,sigma_liq)
        k= k_Green75(Bond)
        
        S = S0  * k**(-2/3)
        Cd= Cd0 / k
    
    M = rho_liq *4/3*np.pi*a0**3
    
    Fv = V**2 - 2*M*g /rho_air /S /Cd
    
    return Fv

def DF_lorenz_tobe_solved_DV(rho_liq,sigma_liq,rho_air,visco,Mair,T,dim_molecule,d0,g,V0,rigid,CdEq):
    """
    Dérivée de l'équation 'F_lorenz_tobe_solved' par rapport à la vitesse 'V'.
    """
    pd = 1e-4
    V1 = V0 * (1-pd)
    V2 = V0 * (1+pd)
    
    Fv_1= F_lorenz_tobe_solved(rho_liq,sigma_liq,rho_air,visco,Mair,T,dim_molecule,d0,g,V1,rigid,CdEq)
    Fv_2= F_lorenz_tobe_solved(rho_liq,sigma_liq,rho_air,visco,Mair,T,dim_molecule,d0,g,V2,rigid,CdEq)
    
    dFdV= (Fv_2-Fv_1)/2/pd/V0
    
    return dFdV

def solve_Vlorenz(rho_liq,sigma_liq,rho_air,visco,Mair,T,dim_molecule,d0,g,V0,rigid,CdEq):
    """
    Résolution par Newton-Raphson de l'équation permettant d'obtenir la vitesse de chute libre des
    gouttes.
    Inputs:
     - rho_liq ------: masse volumique du liquide considéré (kg m^-3).
     - sigma_liq ----: tension superficielle du liquide considéré (N m^-1)
     - rho_air ------: masse volumique de l'air (kg m^-3).
     - visco --------: keyword
                        'GP' ---: the viscosity is given by the law relevant for ideal gases.
                        'LOR' --: the visocity is given by the law use by Lorenz (1993) relevant for N2.
     - Mair ---------: masse molaire des molécules de l'air (kg mol^-1).
     - T ------------: température absolue (K).
     - dim_molecule -: taille des molécules d'air (m).
     - d0 -----------: diamètre de la goutte de liquide, non déformée (m).
     - g ------------: accélération de la pesanteur (m s^-2).
     - V0 -----: première estimation de la vitesse recherchée (m s^-1)
     - rigid --: =0 : gouttes déformées par rapport à la sphère.
                 =1 : gouttes NON-déformées par rapport à la sphère.
     - CdEq ---: type of Cd equation used (cf. Clift p. 111)
    Outputs:
     - Relim --: le nombre de Reynolds correspondant à la vitesse terminale de chute d'une goutte de pluie.
     - Vlim ---: la vitesse terminale de chute d'une goutte de pluie (m s^-1)
    """
    
    # -----------------------------------------------------------
    # On utilise un algorithme de Newton-Raphson : --------------
    precis = 1e-7
    
    Vn  = V0
    Vnp1= 2*Vn
    while abs((Vnp1-Vn)/Vn) > precis:
        Vn   = Vnp1
        Vnp1 = Vn - F_lorenz_tobe_solved(rho_liq,sigma_liq,rho_air,visco,Mair,T,dim_molecule,d0,g,Vn,rigid,CdEq) / \
                    DF_lorenz_tobe_solved_DV(rho_liq,sigma_liq,rho_air,visco,Mair,T,dim_molecule,d0,g,Vn,rigid,CdEq)
        #print (Vn)
    # -----------------------------------------------------------
      
    if visco == 'GP':
        eta_n = visco_gp(Mmol=Mair,T=T,d0=dim_molecule)
    else:
        if visco == 'LOR':
            eta_n = visco_N2_lorenz(T)
        else:
            print (" >>>> WARNING: in function 'F_lorenz_tobe_solved' not relevant viscosity option!!!")
            
    Re_n  = Re_air(rho_air,d0,Vn,eta_n)
    
    return Re_n, Vn

# ===================================================================================================================
# ===================================================================================================================
# ===================================================================================================================
# Afin de pourvoir faire des vérifications, on remet 'Dmax_solid_surf' qui dépend de 'Re_liq' et 'We_liq' elles-aussi
# importées dans ce fichier/module :
# -------------------------------------------------------------------------------------------------------------------
# Reynolds number in the liquid:
def Re_liq(rho_liq,d,U,eta):
    """
    Inputs:
      - rho_liq: liquid density (kg m^-3)
      - d: diameter of the droplet (m)
      - U: terminal velocity (m s^-1)
      - eta: viscosity (Pa s)
    Outputs:
      - Re_liq: Reynolds number in the droplet.
    """
    return rho_liq*d*U/eta

# Weber number in the liquid:
def We_liq(rho_liq,d,U,sigma):
    """
    Inputs:
      - rho_liq: liquid density (kg m^-3)
      - d: diameter of the droplet (m)
      - U: terminal velocity (m s^-1)
      - sigma: surface tension (N m^-1)
    Outputs:
      - We_liq: Weber number in the droplet.
    """
    return rho_liq*d*U**2/sigma

# Extention of the raindrop deformation during an impact with a solid, flat, surface:
def Dmax_solid_surf(rho_liq,rho_air,visco,Mair,T,dim_molecule,g,d,sigma_liq,eta_liq,rigid,CdEq):
    """
    Inputs:
      - rho_liq ---: density of the liquid (kg/m^3).
      - rho_air ---: density of the air (kg/m^3).
      - visco -----: keyword
                        'GP' ---: the viscosity is given by the law relevant for ideal gases.
                        'LOR' --: the visocity is given by the law use by Lorenz (1993) relevant for N2.
      - Mair ------: mean molecular weigth of the air (kg/mol).
      - T ---------: temperature (K).
      - dim_molecule --: size of molecules of the air (m).
      - g ---------: gravity (m/s^2).
      - d ---------: droplet diameter (m).
      - sigma_liq -: surface tension (N m^-1).
      - eta_liq ---: dynamic viscosity of the liquid (Pa s).
      - rigid --------:  =0 : raindrop deformed compared to a sphere.
                         =1 : raindrop NOT-deformed compared to a sphere.
      - CdEq ---------: type of Cd equation used (cf. Clift p. 111)
    Outputs:
      - Dmax ------: diamètre max. d'étalement de la goutte sur une surface plane et solide.
    """
    V0 = 12 # First estimation of the researched raindrops terminal viscosity (m/s). 
    Remax_air, Umax = solve_Vlorenz(rho_liq,sigma_liq,rho_air,visco,Mair,T,dim_molecule,d,g,V0,rigid,CdEq)
    
    #print (Umax)
    We = We_liq(rho_liq,d,Umax,sigma_liq)
    Re = Re_liq(rho_liq,d,Umax,eta_liq)
    
    sqrtP  = np.sqrt( We * Re**(-2/5) )
    A  = 1.24
    
    return Re**(1/5) * sqrtP / (A + sqrtP) #* d
